//
// $Id: Particles.cc 40 2005-06-18 09:07:50Z raphael $
//

#include <GL/freeglut.h>
#include <iostream>
#include <fstream>

#include "Particles.hh"

ParticleSystem::ParticleSystem(Water *water) {
  _water = water;
}

void ParticleSystem::init() {
  GLubyte *buf;
  int size;
  _particles.clear();

  _damping = .03;
  _rate = 2;
  _lifetime = 10;
  _time = 0;
  _timeLast = 0;

  _psize = 1;

  std::ifstream fp;
  fp.open("data/star.bmp", std::ios::in | std::ios::binary);
  fp.seekg(18, std::ios::beg);
  fp.read((char *)&size, 4);

  buf = new GLubyte[size*size*3];

  fp.seekg(54, std::ios::beg);
  fp.read((char *)buf, size*size*3);

  glGenTextures(1, &_texture);
  glBindTexture(GL_TEXTURE_2D, _texture);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, 3, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, buf);

  fp.close();
  delete[] buf;
}

void ParticleSystem::update(float dt) {
  static int ground = 0;

  _time += dt;
  int n = (int)((_time - _timeLast) * _rate);
  _timeLast = ((n > 0) || (_rate == 0)) ? _time : _timeLast;
  for(int i=0;i<n;i++) {
    Particle p;
    p.birth = _time;
    p.pos[0] = 0;
    p.pos[1] = 0;
    p.pos[2] = ground;
    p.v[0] = 16*(((float)rand())/RAND_MAX - .5);
    p.v[1] = 16*(((float)rand())/RAND_MAX - .5);
    p.v[2] = 24*(((float)rand())/RAND_MAX) + 20;
    p.m = .1;
    p.alive = true;
    p.phi = 360*((float)rand())/RAND_MAX;
    _particles.push_back(p);
  }
  
  while(!_particles.empty() &&
        (_time - _particles.front().birth > _lifetime)) {
    _particles.pop_front();
  }
    
  std::deque<Particle>::iterator it = _particles.begin();
  while(it != _particles.end()) {
    if(it->alive) {
      forces(&(*it));
      for(int i=0;i<3;i++) {
        it->v[i] = it->v[i] + dt * it->F[i] / it->m;
        it->pos[i] = it->pos[i] + dt * it->v[i];
      }
      if(it->pos[2] < ground) {
        it->pos[2] = 2*ground - it->pos[2];
        it->v[2] = - it->v[2];
        _water->drop(it->pos);
        it->alive = false;
      }
      it->phi += dt * 100;
      if(it->phi > 360) it->phi -= 360;
    }
    it++;
  }
}

void ParticleSystem::forces(Particle *p) {
  // gravity
  p->F[0] = - _damping * p->v[0];
  p->F[1] = - _damping * p->v[1];
  p->F[2] = -9.81 * p->m - _damping * p->v[2];
}

void ParticleSystem::draw() {
  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glPolygonMode(GL_BACK, GL_FILL);

  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, _texture);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glEnable(GL_BLEND);

  std::deque<Particle>::iterator it = _particles.begin();
  while(it != _particles.end()) {
    if(it->alive) {
      glPushMatrix();
      glTranslatef(it->pos[0], it->pos[2], it->pos[1]);
      glScalef(_psize, _psize, _psize);
      glRotatef(it->phi, 1, .5, 0);
      float c = (_time - it->birth)/_lifetime;
      glColor4f(1-c, .5-(c/2), 0, 1);
      glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(0, 0);
      glVertex3f(-1, 0,-1);
      glTexCoord2f(0, 1);
      glVertex3f(-1, 0, 1);
      glTexCoord2f(1, 0);
      glVertex3f( 1, 0,-1);
      glTexCoord2f(1, 1);
      glVertex3f( 1, 0, 1);
      glEnd(/*GL_TRIANGLE_STRIP*/);
      glColor4f(0, .5-(c/2), 1-c, 1);
      glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(0, 0);
      glVertex3f( 0,-1,-1);
      glTexCoord2f(0, 1);
      glVertex3f( 0,-1, 1);
      glTexCoord2f(1, 0);
      glVertex3f( 0, 1,-1);
      glTexCoord2f(1, 1);
      glVertex3f( 0, 1, 1);
      glEnd(/*GL_TRIANGLE_STRIP*/);
      glColor4f(1-c, 0, .5-(c/2), 1);
      glBegin(GL_TRIANGLE_STRIP);
      glTexCoord2f(0, 0);
      glVertex3f(-1,-1, 0);
      glTexCoord2f(0, 1);
      glVertex3f(-1, 1, 0);
      glTexCoord2f(1, 0);
      glVertex3f( 1,-1, 0);
      glTexCoord2f(1, 1);
      glVertex3f( 1, 1, 0);
      glEnd(/*GL_TRIANGLE_STRIP*/);
      glPopMatrix();
    }
    it++;
  }

  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);

  glPolygonMode(GL_BACK, GL_LINE);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
}

void ParticleSystem::rateInc() {
  _rate += 1;
}

void ParticleSystem::rateDec() {
  _rate -= 1;
  if(_rate < 0) _rate = 0;
}

void ParticleSystem::sizeInc() {
  _psize *= 1.125;
}

void ParticleSystem::sizeDec() {
  _psize /= 1.125;
  if(_psize < .2) _psize = .2;
}
